import io
import paramiko
import pandas as pd
from io import StringIO
import traceback
import logging
import stat
import chardet
import zipfile

from afl_ai_utils.logging_utils import setup_logger, clean_up_logger

logger = setup_logger("afl_ai_utils", log_file="afl_ai_utils.log", level=logging.DEBUG)

class SFTPUtils:
    def __init__(self, host, username, password, port):
        self.host = host
        self.username = username
        self.password = password
        self.port = port

    def list_files_and_dirs(self, remote_dir:str):
        directory_structure = {}
        transport = paramiko.Transport((self.host, self.port))
        transport.connect(username=self.username, password=self.password)
        sftp = paramiko.SFTPClient.from_transport(transport)
        # Get directory structure
        # directory_structure = list_files_and_dirs(sftp, remote_dir)
        for item in sftp.listdir_attr(remote_dir):
            item_path = f"{remote_dir}/{item.filename}"
            # logger.info(directory_structure)
            if stat.S_ISDIR(item.st_mode):
                # Recursive call to list files and dirs in subdirectories
                # print(item_path , "--> ", item.filename)
                directory_structure[item.filename] = self.list_files_and_dirs(remote_dir=item_path)
                # print(directory_structure)
            else:
                if '__files__' not in directory_structure:
                    directory_structure['__files__'] = []
                directory_structure['__files__'].append(item.filename)
        # Close the SFTP session and SSH client
        print(directory_structure)
        sftp.close()
        transport.close()
        return directory_structure

    def print_directory_structure(self, directory_structure: dict, indent=""):
        for key, value in directory_structure.items():
            if key == '__files__':
                for file in value:
                    logger.info(f"{indent}/{file}")
            else:
                print(f"{indent}/{key}")
                self.print_directory_structure(value, indent + f"/{key}")

    def read_from_sftp_server(self, file_path):
        transport = paramiko.Transport((self.host, int(self.port)))
        transport.connect(username=self.username, password=self.password)
        sftp = paramiko.SFTPClient.from_transport(transport)
        df = None

        try:
            if sftp.stat(file_path):
                logger.info(f"Reading the file from SFTP server: {file_path}")

                # Read the file in binary mode
                with sftp.file(file_path, 'rb') as file:
                    file_content = file.read()

                    # Check if the file is a ZIP file
                    if file_path.endswith('.zip'):
                        with zipfile.ZipFile(io.BytesIO(file_content)) as z:
                            # Assuming there's only one file in the ZIP archive
                            zip_info = z.infolist()[0]
                            with z.open(zip_info) as csv_file:
                                df = pd.read_csv(csv_file)
                    else:
                        # Detect the encoding
                        result = chardet.detect(file_content)
                        detected_encoding = result['encoding']
                        logger.info(f"Detected encoding: {detected_encoding}")

                        # Decode the content with the detected encoding
                        decoded_content = file_content.decode(detected_encoding, errors='ignore')
                        buffer = io.StringIO(decoded_content)
                        df = pd.read_csv(buffer)

                    logger.info(f"File read successfully. Number of rows: {len(df)}")
            else:
                logger.info(f"File {file_path} doesn't exist on SFTP server: {file_path}")

        except Exception as e:
            logger.error(f"Error reading file {file_path} from SFTP server: {e}")

        finally:
            sftp.close()
            transport.close()

        return df

    def write_to_sftp_server(self, remote_path: str, dataframe: pd.DataFrame):
        try:
            csv_string = dataframe.to_csv(index=False)
            ssh_client = paramiko.SSHClient()
            ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh_client.connect(self.host, self.port, self.username, self.password)
            sftp = ssh_client.open_sftp()
            csv_buffer = StringIO(csv_string)
            with sftp.file(remote_path, "w") as file:
                file.write(csv_buffer.getvalue())
            sftp.close()
            ssh_client.close()
            logger.info("Data written successfully to", remote_path)
        except Exception as e:
            print("Error:", str(e), "---> ", traceback.format_exc())

    def delete_sftp_file(self, remote_path: str):
        try:
            transport = paramiko.Transport((self.host, self.port))
            transport.connect(username=self.username, password=self.password)
            sftp = paramiko.SFTPClient.from_transport(transport)
            sftp.remove(remote_path)
            sftp.close()
            transport.close()
            logger.info(f"File has been deleted successfully for {remote_path}")
        except Exception as e:
            print("Error:", str(e), "---> ", traceback.format_exc())



    def delete_file_from_sftp_server(self, remote_path: str):
        try:
            ssh_client = paramiko.SSHClient()
            ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh_client.connect(self.host, self.port, self.username, self.password)
            sftp = ssh_client.open_sftp()
            sftp.remove(remote_path)
            sftp.close()
            ssh_client.close()
            logger.info(f"File has been deleted successfully for {remote_path}")
        except Exception as e:
            print("Error:", str(e), "---> ", traceback.format_exc())

    def copy_file_from_sftp_server(self,remote_path:str,dest_path:str):
        try:
            ssh_client = paramiko.SSHClient()
            ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh_client.connect(self.host, self.port, self.username, self.password)
            sftp = ssh_client.open_sftp()
            dest_file_path =[]
            for file in sftp.listdir_attr(remote_path):
                if stat.S_ISREG(file.st_mode):
                    file_path = remote_path + "/" + file.filename
                    local_file = dest_path +"/"+ file.filename
                    print(f"Attempting to copy from {file_path} to {dest_path}")
                    sftp.get(remotepath=file_path, localpath=local_file)
                    dest_file_path.append(file.filename)
            sftp.close()
            ssh_client.close()
            return dest_file_path
        except Exception as e:
            print("Error:", str(e), "---> ", traceback.format_exc())

clean_up_logger(logger=logger)


import os

from dotenv import load_dotenv
from google.cloud import secretmanager
import time

def load_da_secrets():
    from google.cloud import secretmanager
    client = secretmanager.SecretManagerServiceClient()
    secret_name = f"projects/162221169123/secrets/da-secrets/versions/latest"
    response = client.access_secret_version(request={"name": secret_name})
    payload = response.payload.data.decode("UTF-8")
    with open(os.getcwd()+'/.env', 'w', encoding='utf-8') as f:
        f.write(payload)
    load_dotenv()

def load_de_secrets():
    client = secretmanager.SecretManagerServiceClient()
    secret_name = f"projects/162221169123/secrets/de-secrets/versions/latest"
    response = client.access_secret_version(request={"name": secret_name})
    payload = response.payload.data.decode("UTF-8")
    with open('.env', 'w', encoding='utf-8') as f:
        f.write(payload)
    load_dotenv()



import os
import time
import traceback

from google.cloud import bigquery
from google.oauth2.service_account import Credentials
from google.api_core.exceptions import NotFound
import logging
import pandas as pd

logging.basicConfig(format='%(name)s - %(levelname)s -%(asctime)s- %(message)s', level=logging.INFO)


class BigQuery():
    def __init__(self, credential_json=None, scopes=None):
        if credential_json is not None:
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = credential_json

        if scopes == None:
            self.client = bigquery.Client()
        else:
            credentials = Credentials.from_service_account_info(credential_json, scopes=scopes)
            if credential_json is not None:
                self.client = bigquery.Client(credentials=credentials)
            else:
                self.client = bigquery.Client()

    def create_table(self, schema_cols_type: dict, table_id: str, partitioning_field_type: dict,
                     clustering_fields: list):

        schema = []

        for col, col_type in schema_cols_type.items():
            if col_type == "date":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.DATE)
                )
            elif col_type == "integer":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.INT64)
                )

            elif col_type == "float":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.FLOAT64)
                )

            elif col_type == "string":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.STRING)
                )
            elif col_type == "datetime":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.DATETIME)
                )

            elif col_type == "boolean":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.BOOLEAN)
                )
            elif col_type == "timestamp":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.TIMESTAMP)
                )

        table = bigquery.Table(table_id, schema=schema)
        table.clustering_fields = clustering_fields

        if partitioning_field_type and len(partitioning_field_type.keys()) == 1:
            for field, field_type in partitioning_field_type.items():
                if partitioning_field_type[field] == "month":
                    table.time_partitioning = bigquery.TimePartitioning(
                        type_=bigquery.TimePartitioningType.MONTH,
                        field=field
                    )
                elif partitioning_field_type[field] == "day":
                    table.time_partitioning = bigquery.TimePartitioning(
                        type_=bigquery.TimePartitioningType.DAY,
                        field=field
                    )
                elif partitioning_field_type[field] == "integer":
                    table.range_partitioning = bigquery.RangePartitioning(
                        field=field,
                        range_=bigquery.PartitionRange(start=0, end=1000000000, interval=100000),
                    )

        try:
            # Try to get the table
            self.client.get_table(table_id)
            print(f"Table {table_id} already exists.")
        except NotFound:
            # If table is not found, create it
            table = self.client.create_table(table)
            print(
                "Created clustered table {}.{}.{}".format(
                    table.project, table.dataset_id, table.table_id
                )
            )

    def execute_query(self, query: str):
        query_job = self.client.query(query)
        results_df = query_job.result().to_dataframe()  # Waits for job to complete.
        return results_df

    def dump_dataframe_to_bq_table(self, dataframe: pd.DataFrame, schema_cols_type: dict, table_id: str, mode: str):
        """Send a Slack message to a channel via a webhook.

        Args:
            dataframe(pandas dataframe): for dataframe to be dumped to bigquery
            schema_cols_type: {"date_start":"date", "id": "integer", "name": "string"}

            table_id (list): table_id in which dataframe need to be inserted e.g project_id.dataset.table_name = table_id
            mode(str): To append or replace the table - e.g mode = "append"  or mode="replace"
        Returns:
            returns as success message with number of inserted rows and table name
        """
        schema = []
        for col, col_type in schema_cols_type.items():
            if col_type == "date":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.DATE)
                )
            elif col_type == "integer":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.INTEGER)
                )
            elif col_type == "string":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.STRING)
                )
            elif col_type == "datetime":
                schema.append(
                    bigquery.SchemaField(col, bigquery.enums.SqlTypeNames.DATETIME)
                )

        if mode == "replace":
            write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
        else:
            write_disposition = bigquery.WriteDisposition.WRITE_APPEND
        job_config = bigquery.LoadJobConfig(
            # Specify a (partial) schema. All columns are always written to the
            # table. The schema is used to assist in data type definitions.
            schema=schema,
            # Optionally, set the write disposition. BigQuery appends loaded rows
            # to an existing table by default, but with WRITE_TRUNCATE write
            # disposition it replaces the table with the loaded data.
            write_disposition=write_disposition,
        )

        if mode == "append":
            job_config.schema_update_options = [
                bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION
            ]

        load_job = None
        retry_count = 0
        exception = None
        while load_job is None and retry_count < 5:
            try:
                load_job = self.client.load_table_from_dataframe(
                    dataframe, table_id, job_config=job_config
                )  # Make an API request.
                load_job.result()  # Wait for the job to complete.

                table = self.client.get_table(table_id)  # Make an API request.
                message = "Loaded {} rows and {} columns to {} ".format(load_job.output_rows,
                                                                        len(table.schema),
                                                                        table_id)
                logging.info(message)
                return message
            except Exception as e:
                exception = e
                retry_count += 1
                print(traceback.format_exc())
                time.sleep(5)
        raise Exception(f"BigQuery load job given exception in all the {retry_count} retries:  {exception}")

    def create_or_update_view(self, table_path, view_src):
        # Define the view configuration
        table = bigquery.Table(table_path)
        table.view_query = view_src
        try:
            existing_view = self.client.get_table(table)  # Try fetching the view
            logging.info(f"Updating existing view: {table}")
        except Exception:
            logging.info(f"Creating new view: {table}")

        # Create or update the view
        view = self.client.update_table(table, ["view_query"]) if "existing_view" in locals() else self.client.create_table(
            table)

        logging.info(f"View {table} has been created/updated successfully!")


import setuptools

with open("README.md", "r") as fh:
    description = fh.read()

setuptools.setup(
    name="afl-ai-utils",
    version="0.6.4",
    author="Abhay Kumar",
    author_email="abhay.kumar@arvindbrands.co.in",
    packages=["afl_ai_utils"],
    description="A sample test package",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/gituser/test-tackage",
    license='MIT',
    python_requires='>=3.8',
    install_requires=["google-cloud-secret-manager","requests", "google-cloud-bigquery", "pandas", "db_dtypes", "fake_headers", "html5lib", "webdriver-manager"]
)


import os
import os.path
import os.path
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
import io
from googleapiclient.discovery import build
from googleapiclient.http import MediaIoBaseDownload
import pandas as pd
import requests
import json


# If modifying these scopes, delete the file token.json.

def get_drive_token(client_id, client_secret, refresh_token):
    url = "https://login.microsoftonline.com/d6454b9f-4ca2-4392-b62f-20e21e54335a/oauth2/v2.0/token"

    payload = f'client_id={client_id}&scope=offline_access%20Mail.ReadWrite%20Mail.send&grant_type=refresh_token&client_secret={client_secret}&redirect_uri=https%3A%2F%2Flocalhost&refresh_token={refresh_token}'
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Cookie': 'buid=0.AXIAn0tF1qJMkkO2LyDiHlQzWv8X8zrAVMtPqebAWud5G3tyAAA.AQABAAEAAAAmoFfGtYxvRrNriQdPKIZ-kZ5EkppMdTFRuYfniN_3y8-Xd4UdnOxoj73wJ1eZZCaKAT8JgDNpeqo0oFp59_urEfHRgAmxblfrvqZtteL87F0uVuBqo0iR-vyvN064OsAgAA; esctx=PAQABAAEAAAAmoFfGtYxvRrNriQdPKIZ-P8XcBu5-Z0yzoAJsYjHRoOrKnytCJJQZgtooKwI6-pwI0SH2MDCLXDXZbbdMx66FO30bZMF3OHeT5bL1TAdiQ4VV253aXWPOxep_3bDQ51hSp_8t3O5-_onUpGnU8RcuxnipsDXB1JojtTrJv8z5wQjGcGTbcAT1ttYODVqDMgIgAA; esctx-FbQVuPalhqg=AQABAAEAAAAmoFfGtYxvRrNriQdPKIZ-3CuntieqvcMk6UUaMPDkg0albmtVMGDKcYBLElzQlZY5BuS8kjO13PYFuvM631jlMLKFAtIxhAhe-6ffmXxJ7FIvp6zkgdlCdjY3zoSgYxAmMXLDc8II0tV8QHkyO8MN_C-kClYgVW8qHbel_H_FjCAA; fpc=Am86eFjVZTxGmw3t5Z7iqnHZjOJlAQAAAAuHAt0OAAAAlaghQQEAAABQhwLdDgAAAP8yaQoCAAAA5IgC3Q4AAAA; stsservicecookie=estsfd; x-ms-gateway-slice=estsfd'
    }

    response = requests.request("POST", url, headers=headers, data=payload)
    result = json.loads(response.text)
    return result['access_token']


def get_files_id_name(credential_json_file, token_json_file):
    """Shows basic usage of the Drive v3 API.
  Prints the names and ids of the first 10 files the user has access to.
  """
    SCOPES = ["https://www.googleapis.com/auth/drive.metadata.readonly", 'https://www.googleapis.com/auth/drive']
    creds = None
    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    if os.path.exists(token_json_file):
        creds = Credentials.from_authorized_user_file(token_json_file, SCOPES)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                credential_json_file, SCOPES
            )
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(token_json_file, "w") as token:
            token.write(creds.to_json())

    service = build("drive", "v3", credentials=creds)

    # Call the Drive v3 API
    results = (
        service.files()
        .list(pageSize=10, fields="nextPageToken, files(id, name)")
        .execute()
    )
    items = results.get("files", [])

    if not items:
        print("No files found.")
        return
    file_id_name = {}
    for item in items:
        file_id_name[item['id']] = item['name']
    return file_id_name


def download_drive_file(file_id, file_name, credential_json_file, token_json_file):
    """Shows basic usage of the Drive v3 API.
  Prints the names and ids of the first 10 files the user has access to.
  """
    SCOPES = ['https://www.googleapis.com/auth/cloud-platform', 'https://www.googleapis.com/auth/drive',
              "https://www.googleapis.com/auth/bigquery"]

    creds = None
    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    if os.path.exists(token_json_file):
        creds = Credentials.from_authorized_user_file(token_json_file, SCOPES)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                credential_json_file, SCOPES
            )
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(token_json_file, "w") as token:
            token.write(creds.to_json())

    service = build("drive", "v3", credentials=creds)

    file_id = file_id

    request = service.files().get_media(fileId=file_id)
    file = io.FileIO(file_name, mode='wb')
    downloader = MediaIoBaseDownload(file, request)
    done = False
    while done is False:
        status, done = downloader.next_chunk()
        print(f"Download {int(status.progress() * 100)}.")
    return True
